from io import BytesIO
from unittest.mock import patch

from openpyxl import Workbook
import pandas as pd
import pytest

from viadot.exceptions import CredentialError
from viadot.sources import Sharepoint
from viadot.sources.sharepoint import SharepointCredentials


DUMMY_CREDS = {
    "site": "tenant.sharepoint.com",  # pragma: allowlist secret
    "client_id": "dummy_client_id",  # pragma: allowlist secret
    "tenant_id": "dummy_tenant_id",  # pragma: allowlist secret
    "client_secret": "dummy_client_secret",  # pragma: allowlist secret
}
SAMPLE_DF = pd.DataFrame(
    {
        "int_col": [1, 2, 3, 4, 5, None],
        "float_col": [1.1, 2.2, 3.3, 3.0, 5.5, 6.6],
        "str_col": ["a", "b", "c", "d", "e", "f"],
        "nan_col": [None, None, None, None, None, None],
        "mixed_col": [1, "text", None, None, 4.2, "text2"],
    }
)


def create_excel_file():
    """Create a simple Excel file in memory for testing."""
    wb = Workbook()
    ws = wb.active
    ws.title = "Sheet1"
    ws["A1"] = "col_a"
    ws["B1"] = "col_b"
    ws["A2"] = "val1"
    ws["B2"] = "val1"
    ws["A3"] = ""
    ws["B3"] = "val2"
    ws["A4"] = "val2"
    ws["B4"] = "val3"
    ws["A5"] = "NA"
    ws["B5"] = "val4"
    ws["A6"] = "N/A"
    ws["B6"] = "val5"
    ws["A7"] = "#N/A"
    ws["B7"] = "val6"
    bio = BytesIO()
    wb.save(bio)
    bio.seek(0)
    return bio.getvalue()


class SharepointMock(Sharepoint):
    def _download_file_stream(self, _url: str | None = None, **kwargs):
        if "nrows" in kwargs:
            msg = "Parameter 'nrows' is not supported."
            raise ValueError(msg)

        return pd.ExcelFile(BytesIO(create_excel_file()))


@pytest.fixture
def sharepoint_mock():
    return SharepointMock(credentials=DUMMY_CREDS)


@pytest.fixture
def sharepoint():
    credentials = {
        "site": "example.sharepoint.com",  # pragma: allowlist secret
        "client_id": "dummy_client_id",  # pragma: allowlist secret
        "tenant_id": "dummy_tenant_id",  # pragma: allowlist secret
        "client_secret": "dummy_client_secret",  # pragma: allowlist secret
    }
    return Sharepoint(credentials=credentials)


def test_valid_credentials():
    credentials = {
        "site": "tenant.sharepoint.com",  # pragma: allowlist secret
        "client_id": "client",  # pragma: allowlist secret
        "tenant_id": "tenant",  # pragma: allowlist secret
        "client_secret": "dummy_client_secret",  # pragma: allowlist secret
    }
    shrp_creds = SharepointCredentials(**credentials)
    assert shrp_creds.site == credentials["site"]
    assert shrp_creds.client_id == credentials["client_id"]
    assert shrp_creds.tenant_id == credentials["tenant_id"]
    assert shrp_creds.client_secret == credentials["client_secret"]


def test_invalid_authentication():
    credentials = {
        "site": "tenant.sharepoint.com",  # pragma: allowlist secret
        "client_id": "client",  # pragma: allowlist secret
        "client_secret": "dummy_client_secret",  # pragma: allowlist secret
        "tenant_id": "tenant",  # pragma: allowlist secret
    }

    s = Sharepoint(credentials=credentials)

    # Patch the GraphClient constructor to simulate an authentication failure
    with patch("viadot.sources.sharepoint.GraphClient") as graph_client:
        graph_client.side_effect = Exception("Authentication failed")

        with pytest.raises(
            CredentialError,
            match="Could not authenticate to tenant.sharepoint.com with provided credentials.",
        ):
            s.get_client()


def test_missing_client_id():
    credentials = {
        "site": "example.sharepoint.com",  # pragma: allowlist secret
        "tenant_id": "dummy_tenant_id",  # pragma: allowlist secret
        "client_secret": "dummy_client_secret",  # pragma: allowlist secret
    }
    with pytest.raises(
        CredentialError,
        match=r"Missing required credentials: \['client_id'\]",
    ):
        SharepointCredentials(**credentials)


def test_acquire_token_with_client_secret():
    creds = {
        "site": "tenant.sharepoint.com",  # pragma: allowlist secret
        "client_id": "client_id",  # pragma: allowlist secret
        "tenant_id": "tenant_id",  # pragma: allowlist secret
        "client_secret": "super_secret",  # pragma: allowlist secret
    }
    sp = Sharepoint(credentials=creds)

    captured = {}

    class _DummyApp:
        def __init__(self, authority, client_id, client_credential):
            captured["authority"] = authority
            captured["client_id"] = client_id
            captured["client_credential"] = client_credential

        def acquire_token_for_client(self, scopes=None, **_kwargs):
            captured["scopes"] = scopes
            return {"access_token": "token"}

    with patch(
        "viadot.sources.sharepoint.msal.ConfidentialClientApplication", new=_DummyApp
    ):
        token = sp._acquire_token_func()

    assert isinstance(token, dict)
    assert captured["client_id"] == creds["client_id"]
    assert creds["tenant_id"] in captured["authority"]
    # In client_secret flow, msal should receive the secret string as client_credential
    assert captured["client_credential"] == creds["client_secret"]


def test_acquire_token_with_certificate():
    # pass binary certificate via `credentials` and the rest via `credentials_cert_auth`
    mocked_cert_auth = {
        "site": "tenant.sharepoint.com",  # pragma: allowlist secret
        "client_id": "dummy_client_id",  # pragma: allowlist secret
        "tenant_id": "dummy_tenant_id",  # pragma: allowlist secret
        "certificate_password": "dummy_certificate_password",  # pragma: allowlist secret
        "certificate": b"blah_blah",
    }

    captured = {}

    class _DummyApp:
        def __init__(self, authority, client_id, client_credential):
            captured["authority"] = authority
            captured["client_id"] = client_id
            captured["client_credential"] = client_credential

        def acquire_token_for_client(self, scopes=None, **_kwargs):
            captured["scopes"] = scopes
            return {"access_token": "token"}

    with patch(
        "viadot.sources.sharepoint.msal.ConfidentialClientApplication",
        new=_DummyApp,
    ):
        sp = Sharepoint(
            credentials=mocked_cert_auth,
        )
        token = sp._acquire_token_func()

    # Token returned
    assert isinstance(token, dict)
    # MSAL was initialized with expected values
    assert captured["client_id"] == mocked_cert_auth["client_id"]
    assert mocked_cert_auth["tenant_id"] in captured["authority"]
    # Certificate flow uses a dict client_credential with passphrase and path
    assert isinstance(captured["client_credential"], dict)
    assert (
        captured["client_credential"]["passphrase"]
        == mocked_cert_auth["certificate_password"]
    )
    assert "private_key_pfx_path" in captured["client_credential"]


def test_sharepoint_default_na(sharepoint_mock):
    df = sharepoint_mock.to_df(
        url="test/file.xlsx", na_values=Sharepoint.DEFAULT_NA_VALUES
    )

    assert not df.empty
    assert "NA" not in list(df["col_a"])


def test_sharepoint_custom_na(sharepoint_mock):
    df = sharepoint_mock.to_df(
        url="test/file.xlsx",
        na_values=[v for v in Sharepoint.DEFAULT_NA_VALUES if v != "NA"],
    )

    assert not df.empty
    assert "NA" in list(df["col_a"])


def test__get_file_extension(sharepoint_mock):
    url_excel = "https://tenant.sharepoint.com/sites/site/file.xlsx"
    url_dir = "https://tenant.sharepoint.com/sites/site/"
    url_txt = "https://tenant.sharepoint.com/sites/site/file.txt"

    excel_ext = sharepoint_mock._get_file_extension(url=url_excel)
    txt_ext = sharepoint_mock._get_file_extension(url=url_txt)
    dir_ext = sharepoint_mock._get_file_extension(url=url_dir)

    assert excel_ext == ".xlsx"
    assert txt_ext == ".txt"
    assert dir_ext == ""


def test__is_file(sharepoint_mock):
    is_file = sharepoint_mock._is_file(url="https://example.com/file.xlsx")
    assert is_file is True

    is_file = sharepoint_mock._is_file(url="https://example.com/dir")
    assert is_file is False


def test__parse_excel_single_sheet(sharepoint_mock):
    excel_file = sharepoint_mock._download_file_stream()
    result_df = sharepoint_mock._parse_excel(excel_file, sheet_name="Sheet1")
    expected = pd.DataFrame(
        {
            "col_a": ["val1", "", "val2", "NA", "N/A", "#N/A"],
            "col_b": ["val1", "val2", "val3", "val4", "val5", "val6"],
        }
    )

    assert result_df["col_b"].equals(expected["col_b"])


def test__parse_excel_string_dtypes(sharepoint_mock):
    excel_file = sharepoint_mock._download_file_stream()
    result_df = sharepoint_mock._parse_excel(excel_file, sheet_name="Sheet1")

    for column in result_df.columns:
        assert result_df[column].dtype == object


def test__load_and_parse_not_valid_extension(sharepoint_mock):
    with pytest.raises(ValueError):  # noqa: PT011
        sharepoint_mock._load_and_parse(file_url="https://example.com/file.txt")


def test_scan_sharepoint_folder_valid_url(sharepoint_mock):
    url = "https://company.sharepoint.com/sites/site_name/final_folder/"

    # Build a fake Graph client hierarchy to return children with names
    class _FakeChildren:
        def get(self):
            class _Exec:
                def execute_query(self):
                    class _Item:
                        def __init__(self, name):
                            self.name = name

                    return [_Item("file1.txt"), _Item("file2.txt")]

            return _Exec()

    class _FakeDriveItem:
        def get(self):
            class _Exec:
                def execute_query(self):
                    # Return an object that has .children
                    class _FolderItem:
                        children = _FakeChildren()

                    return _FolderItem()

            return _Exec()

    class _FakeByUrl:
        @property
        def drive_item(self):
            return _FakeDriveItem()

    class _FakeShares:
        def by_url(self, _url):
            return _FakeByUrl()

    class _FakeClient:
        shares = _FakeShares()

    with patch.object(Sharepoint, "get_client", return_value=_FakeClient()):
        expected_files = [
            (
                "https://company.sharepoint.com/sites/site_name/"
                "final_folder/file1.txt"
            ),
            (
                "https://company.sharepoint.com/sites/site_name/"
                "final_folder/file2.txt"
            ),
        ]

        result = sharepoint_mock.scan_sharepoint_folder(url)
        assert result == expected_files


def test_scan_sharepoint_folder_invalid_url(sharepoint_mock):
    url = "https://company.sharepoint.com/folder/sub_folder/final_folder"

    with pytest.raises(ValueError, match="URL does not contain '/sites/' segment."):
        sharepoint_mock.scan_sharepoint_folder(url)


def test_scan_sharepoint_folder_empty_response(sharepoint_mock):
    url = (
        "https://company.sharepoint.com/sites/site_name/folder/sub_folder/final_folder"
    )

    # Fake client that returns no children
    class _FakeChildren:
        def get(self):
            class _Exec:
                def execute_query(self):
                    return []

            return _Exec()

    class _FakeDriveItem:
        def get(self):
            class _Exec:
                def execute_query(self):
                    class _FolderItem:
                        children = _FakeChildren()

                    return _FolderItem()

            return _Exec()

    class _FakeByUrl:
        @property
        def drive_item(self):
            return _FakeDriveItem()

    class _FakeShares:
        def by_url(self, _url):
            return _FakeByUrl()

    class _FakeClient:
        shares = _FakeShares()

    with patch.object(Sharepoint, "get_client", return_value=_FakeClient()):
        result = sharepoint_mock.scan_sharepoint_folder(url)
        assert result == []


def test_download_file_stream_unsupported_param(sharepoint_mock):
    url = "https://company.sharepoint.com/sites/site_name/folder/test_file.xlsx"

    with pytest.raises(ValueError, match="Parameter 'nrows' is not supported."):
        sharepoint_mock._download_file_stream(url, nrows=10)


def test_successful_download(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/file.xlsx"
    with patch.object(
        Sharepoint,
        "_download_file_stream",
        return_value=pd.ExcelFile(BytesIO(create_excel_file())),
    ):
        df = sharepoint.to_df(url)
        assert not df.empty
        assert len(df) == 6


def test_access_denied(sharepoint):
    url = (
        "https://example.sharepoint.com/sites/site/"
        "Shared%20Documents/restricted_file.xlsx"
    )
    with (
        patch.object(
            Sharepoint,
            "_download_file_stream",
            side_effect=PermissionError("403 Forbidden"),
        ),
        pytest.raises(PermissionError),
    ):
        sharepoint.to_df(url)


def test_invalid_excel_file(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/file.xlsx"
    with (
        patch.object(
            Sharepoint,
            "_download_file_stream",
            side_effect=ValueError("Excel file format cannot be determined"),
        ),
        pytest.raises(ValueError, match="Excel file format cannot be determined"),
    ):
        sharepoint.to_df(url)


def test_mixed_file_extensions(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/folder"
    with (
        patch.object(Sharepoint, "scan_sharepoint_folder") as mock_scan,
        patch.object(Sharepoint, "_load_and_parse") as mock_load_and_parse,
    ):
        mock_scan.return_value = [
            url + "/file1.xlsx",
            url + "/file2.txt",
            url + "/file3.pdf",
        ]
        mock_load_and_parse.return_value = pd.DataFrame(
            {
                "col_a": ["val1", "", "val2", "NA", "N/A", "#N/A"],
                "col_b": ["val1", "val2", "val3", "val4", "val5", "val6"],
            }
        )
        df = sharepoint.to_df(url)

        # Verify that _load_and_parse was only called for .xlsx files
        mock_load_and_parse.assert_called_once_with(
            file_url=url + "/file1.xlsx",
            sheet_name=None,
            na_values=None,
            **{},
        )

        assert not df.empty
        assert len(df) == 6


def test_empty_folder(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/folder"
    with (
        patch.object(Sharepoint, "scan_sharepoint_folder") as mock_scan,
    ):
        mock_scan.return_value = []
        df = sharepoint.to_df(url)
        assert df.empty



"""Flow for downloading the data from Superpetrics and uploading it to ADLS."""

from typing import Any

from prefect import flow
from prefect.task_runners import ConcurrentTaskRunner

from viadot.orchestration.prefect.tasks import (
    df_to_adls,
    supermetrics_to_df,
)


@flow(
    name="Supermetrics extraction to ADLS",
    description="Extract data from Supermetrics and load it into ADLS.",
    retries=1,
    retry_delay_seconds=60,
    task_runner=ConcurrentTaskRunner,
)
def supermetrics_to_adls(
    # supermetrics
    query_params: dict[str, Any] | None = None,
    # ADLS
    adls_path: str | None = None,
    overwrite: bool = False,
    # Auth
    supermetrics_credentials_secret: str | None = None,
    supermetrics_config_key: str | None = None,
    adls_credentials_secret: str | None = None,
    adls_config_key: str | None = None,
    **kwargs: dict[str, Any] | None,
) -> None:
    """Flow to extract data from the Supermetrics API and save it to ADLS.

    This function queries data from the Supermetrics API using the provided query
    parameters and saves the resulting DataFrame to Azure Data Lake Storage (ADLS)
    as a file.

    Args:
    ----
        query_params (dict[str, Any], optional):
            A dictionary of query parameters for the Supermetrics API. These parameters
            specify the data to retrieve from Supermetrics. If not provided, the default
            parameters from the Supermetrics configuration will be used.
        adls_path (str, optional):
            The destination path in ADLS where the DataFrame will be saved. This should
            include the file name and extension (e.g., 'myfolder/myfile.csv'). If not
            provided, the function will use a default path from the configuration
            or raise an error.
        overwrite (bool, optional):
            A flag indicating whether to overwrite the existing file in ADLS. If set
            to Falseand the file exists, an error will be raised. Default is False.
        supermetrics_credentials_secret (str, optional):
            The name of the secret in the secret management system containing
            the Supermetrics API credentials. If not provided, the function will use
            credentials specified in the configuration.
        supermetrics_config_key (str, optional):
            The key in the viadot configuration holding relevant credentials.
            Defaults to None.
        adls_credentials_secret (str, optional):
            The name of the secret in the secret management system containing
            the ADLS credentials. If not provided, the function will use credentials
            specified in the configuration.
        adls_config_key (str, optional):
            The key in the viadot configuration holding relevant credentials.
            Defaults to None.
        **kwargs (dict[str, Any], optional):
            Additional keyword arguments to pass to the `supermetrics_to_df` function
            for further customization of the Supermetrics query.

    Raises:
    ------
        ValueError:
            If `adls_path` is not provided and cannot be determined from
            the configuration.

    """
    df = supermetrics_to_df(
        query_params=query_params,
        credentials_secret=supermetrics_credentials_secret,
        config_key=supermetrics_config_key,
        **kwargs,
    )

    return df_to_adls(
        df=df,
        path=adls_path,
        credentials_secret=adls_credentials_secret,
        config_key=adls_config_key,
        overwrite=overwrite,
    )



"""A module for interacting with DuckDB."""

import re
from typing import Literal

import duckdb
import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.signals import SKIP
from viadot.sources.base import Record, Source


class DuckDBCredentials(BaseModel):
    database: str
    read_only: bool = True


class DuckDB(Source):
    DEFAULT_SCHEMA = "main"

    def __init__(
        self,
        config_key: str | None = None,
        credentials: DuckDBCredentials | None = None,
        *args,
        **kwargs,
    ):
        """A class for interacting with DuckDB.

        Args:
            config_key (str, optional): The key inside local config containing the
                credentials.
            credentials (DuckDBCredentials, optional): Credentials for the connection.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to None.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        if credentials is None:
            msg = "Please specify the credentials."
            raise CredentialError(msg)
        validated_creds = dict(
            DuckDBCredentials(**raw_creds)
        )  # validate the credentials
        super().__init__(*args, credentials=validated_creds, **kwargs)

    @property
    def con(self) -> duckdb.DuckDBPyConnection:
        """Return a new connection to the database.

        As the views are highly isolated, we need a new connection for each query in
        order to see the changes from previous queries (eg. if we create a new table and
        then we want to list tables from INFORMATION_SCHEMA, we need to create a new
        DuckDB connection).

        Returns:
            duckdb.DuckDBPyConnection: database connection.
        """
        return duckdb.connect(
            database=self.credentials.get("database"),
            read_only=self.credentials.get("read_only", False),
        )

    @property
    def tables(self) -> list[str]:
        """Show the list of fully qualified table names.

        Returns:
            list[str]: The list of tables in the format '{SCHEMA}.{TABLE}'.
        """
        tables_meta: list[tuple] = self.run_query(
            "SELECT * FROM information_schema.tables"
        )
        return [table_meta[1] + "." + table_meta[2] for table_meta in tables_meta]

    @property
    def schemas(self) -> list[str]:
        """Show the list of schemas.

        Returns:
            list[str]: The list of schemas.
        """
        self.logger.warning(
            "DuckDB does not expose a way to list schemas. `DuckDB.schemas` only contains schemas with tables."
        )
        tables_meta: list[tuple] = self.run_query(
            "SELECT * FROM information_schema.tables"
        )
        return [table_meta[1] for table_meta in tables_meta]

    def to_df(self, query: str, if_empty: str | None = None) -> pd.DataFrame:
        """Run DuckDB query and save output to a pandas DataFrame.

        Args:
            query (str): The query to execute. If query doesn't start with SELECT or
                WITH, empty DataFrame will be returned.
            if_empty (str, optional): What to do if output DataFrame is empty. Defaults
                to None.

        Returns:
            pd.DataFrame: DataFrame with query output
        """
        if query.upper().startswith("SELECT") or query.upper().startswith("WITH"):
            df = self.run_query(query, fetch_type="dataframe")
            if df.empty:
                self._handle_if_empty(if_empty=if_empty)
        else:
            df = pd.DataFrame()
        return df

    def run_query(
        self, query: str, fetch_type: Literal["record", "dataframe"] = "record"
    ) -> list[Record] | bool:
        """Run a query on DuckDB.

        Args:
            query (str): The query to execute.
            fetch_type (Literal[, optional): How to return the data: either in the
                default record format or as a pandas DataFrame. Defaults to "record".

        Returns:
            Union[list[Record], bool]: Either the result set of a query or,
                in case of DDL/DML queries, a boolean describing whether
                the query was executed successfully.
        """
        allowed_fetch_type_values = ["record", "dataframe"]
        if fetch_type not in allowed_fetch_type_values:
            msg = f"Only the values {allowed_fetch_type_values} are allowed for 'fetch_type'"
            raise ValueError(msg)
        cursor = self.con.cursor()
        cursor.execute(query)

        # Cleanup the query.
        query_clean = query.upper().strip()
        # Find comments.
        regex = r"^\s*[--;].*"
        lines = query_clean.splitlines()
        final_query = ""

        for line_raw in lines:
            line = line_raw.strip()
            match_object = re.match(regex, line)
            if not match_object:
                final_query += " " + line
        final_query = final_query.strip()
        query_keywords = ["SELECT", "SHOW", "PRAGMA", "WITH"]
        if any(final_query.startswith(word) for word in query_keywords):
            result = cursor.fetchall() if fetch_type == "record" else cursor.fetchdf()
        else:
            result = True

        cursor.close()
        return result

    def _handle_if_empty(self, if_empty: str = "warn") -> None:
        if if_empty == "warn":
            self.logger.warning("The query produced no data.")
        elif if_empty == "skip":
            msg = "The query produced no data. Skipping..."
            raise SKIP(msg)
        elif if_empty == "fail":
            msg = "The query produced no data."
            raise ValueError(msg)

    def create_table_from_parquet(
        self,
        table: str,
        path: str,
        schema: str | None = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = "fail",
    ) -> bool:
        """Create a DuckDB table with a CTAS from Parquet file(s).

        Args:
            table (str): Destination table.
            path (str): The path to the source Parquet file(s). Glob expressions are
                also allowed here (eg. `my_folder/*.parquet`).
            schema (str, optional): Destination schema. Defaults to None.
            if_exists (Literal[, optional): What to do if the table already exists.
            The 'delete' option deletes data and then inserts new one. Defaults to
                "fail".

        Raises:
            ValueError: If the table exists and `if_exists` is set to `fail`.

        Returns:
            None: Does not return anything.
        """
        schema = schema or DuckDB.DEFAULT_SCHEMA
        fqn = schema + "." + table
        exists = self._check_if_table_exists(schema=schema, table=table)

        if exists:
            if if_exists == "replace":
                self.run_query(f"DROP TABLE {fqn}")
            elif if_exists == "append":
                self.logger.info(f"Appending to table {fqn}...")
                create_table_query = f"COPY {fqn} FROM '{path}' (FORMAT 'parquet')"
                self.run_query(create_table_query)
                self.logger.info(f"Successfully appended data to table '{fqn}'.")
                return True
            elif if_exists == "delete":
                self.run_query(f"DELETE FROM {fqn}")  # noqa: S608
                self.logger.info(f"Successfully deleted data from table '{fqn}'.")
                self.run_query(
                    f"INSERT INTO {fqn} SELECT * FROM read_parquet('{path}')"  # noqa: S608
                )
                self.logger.info(f"Successfully inserted data into table '{fqn}'.")
                return True
            elif if_exists == "fail":
                msg = "The table already exists and 'if_exists' is set to 'fail'."
                raise ValueError(msg)
            elif if_exists == "skip":
                return False
        self.run_query(f"CREATE SCHEMA IF NOT EXISTS {schema}")
        self.logger.info(f"Creating table {fqn}...")
        create_table_query = f"CREATE TABLE {fqn} AS SELECT * FROM '{path}';"  # noqa: S608
        self.run_query(create_table_query)
        self.logger.info(f"Table {fqn} has been created successfully.")
        return True

    def drop_table(self, table: str, schema: str | None = None) -> bool:
        """Drop a table.

        A thin wrapper around DuckDB.run_query(), with additional logs.

        Args:
            table (str): The table to be dropped.
            schema (str, optional): The schema where the table is located. Defaults to
                None.

        Returns:
            bool: Whether the table was dropped.
        """
        schema = schema or DuckDB.DEFAULT_SCHEMA
        fqn = schema + "." + table

        self.logger.info(f"Dropping table {fqn}...")
        dropped = self.run_query(f"DROP TABLE IF EXISTS {fqn}")
        if dropped:
            self.logger.info(f"Table {fqn} has been dropped successfully.")
        else:
            self.logger.info(f"Table {fqn} could not be dropped.")
        return dropped

    def _check_if_table_exists(self, table: str, schema: str | None = None) -> bool:
        schema = schema or DuckDB.DEFAULT_SCHEMA
        fqn = schema + "." + table
        return fqn in self.tables

    def _check_if_schema_exists(self, schema: str) -> bool:
        if schema == DuckDB.DEFAULT_SCHEMA:
            return True
        fqns = self.tables
        return any(fqn.split(".")[0] == schema for fqn in fqns)



from pathlib import Path
import shutil

from loguru import logger

from viadot.orchestration.prefect.tasks import clone_repo


def test_clone_repo_private(AZURE_REPO_URL):
    test_repo_dir = "test_repo_dir"

    assert not Path(test_repo_dir).exists()

    clone_repo.fn(url=AZURE_REPO_URL, path=test_repo_dir, logger=logger)

    assert Path(test_repo_dir).exists()

    shutil.rmtree(test_repo_dir)



"""'mediatool.py'."""

import json
from typing import Literal

import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import APIError, CredentialError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_response, join_dfs


class MediatoolCredentials(BaseModel):
    """Checking for values in Mediatool credentials dictionary.

    Two key values are held in the Mediatool connector:
        - user_id: The unique ID for the organization.
        token: A unique token to be used as the password for API requests.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    user_id: str
    token: str


class Mediatool(Source):
    """Class implementing the Mediatool API.

    Download data from Mediatool platform. Using Mediatool class user is able to
    download organizations, media entries, campaigns, vehicles, and media types data.
    """

    def __init__(
        self,
        *args,
        credentials: MediatoolCredentials | None = None,
        config_key: str | None = None,
        user_id: str | None = None,
        **kwargs,
    ):
        """Create an instance of the Mediatool class.

        Args:
            credentials (MediatoolCredentials, optional): Meditaool credentials.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to None.
            user_id (str, optional): User ID. Defaults to None.
        """
        credentials = credentials or get_source_credentials(config_key) or None
        if credentials is None:
            message = "Missing credentials."
            raise CredentialError(message)

        validated_creds = dict(MediatoolCredentials(**credentials))
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.header = {"Authorization": f"Bearer {credentials.get('token')}"}
        self.user_id = user_id or credentials.get("user_id")

    def _get_organizations(
        self,
        user_id: str | None = None,
    ) -> list[dict[str, str]]:
        """Get organizations data based on the user ID.

        Args:
            user_id (str, optional): User ID. Defaults to None.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        user_id = user_id or self.user_id
        url_organizations = f"https://api.mediatool.com/users/{user_id}/organizations"

        response = handle_api_response(
            url=url_organizations,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)
        organizations = response_dict["organizations"]

        list_organizations = []
        for org in organizations:
            list_organizations.append(
                {
                    "_id": org.get("_id"),
                    "name": org.get("name"),
                    "abbreviation": org.get("abbreviation"),
                }
            )

        return list_organizations

    def _get_media_entries(
        self,
        organization_id: str,
    ) -> list[dict[str, str]]:
        """Data for media entries.

        Args:
            organization_id (str): Organization ID.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        url = (
            "https://api.mediatool.com/searchmediaentries?q="
            + f'{{"organizationId": "{organization_id}"}}'
        )

        response = handle_api_response(
            url=url,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)

        return response_dict["mediaEntries"]

    def _get_vehicles(
        self,
        vehicle_ids: list[str],
    ) -> list[dict[str, str]]:
        """Vehicles data based on the organization IDs.

        Args:
            vehicle_ids (list[str]): List of organization IDs.

        Raises:
            APIError: Mediatool API does not recognise the vehicle id.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        response_list = []
        missing_vehicles = []

        for vid in vehicle_ids:
            url = f"https://api.mediatool.com/vehicles/{vid}"
            try:
                response = handle_api_response(
                    url=url,
                    headers=self.header,
                    method="GET",
                )
            except APIError:
                missing_vehicles.append(vid)
            else:
                response_dict = json.loads(response.text)
                response_list.append(response_dict["vehicle"])

        if missing_vehicles:
            self.logger.error(f"Vehicle were not found for: {missing_vehicles}.")

        return response_list

    def _get_campaigns(
        self,
        organization_id: str,
    ) -> list[dict[str, str]]:
        """Campaign data based on the organization ID.

        Args:
            organization_id (str): Organization ID.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        url_campaigns = (
            f"https://api.mediatool.com/organizations/{organization_id}/campaigns"
        )

        response = handle_api_response(
            url=url_campaigns,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)

        return response_dict["campaigns"]

    def _get_media_types(
        self,
        media_type_ids: list[str],
    ) -> list[dict[str, str]]:
        """Media types data based on the media types ID.

        Args:
            media_type_ids (list[str]): List of media type IDs.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        list_media_types = []
        for id_media_type in media_type_ids:
            response = handle_api_response(
                url=f"https://api.mediatool.com/mediatypes/{id_media_type}",
                headers=self.header,
                method="GET",
            )
            response_dict = json.loads(response.text)
            list_media_types.append(
                {
                    "_id": response_dict.get("mediaType").get("_id"),
                    "name": response_dict.get("mediaType").get("name"),
                    "type": response_dict.get("mediaType").get("type"),
                }
            )

        return list_media_types

    def _to_records(
        self,
        endpoint: Literal[
            "organizations", "media_entries", "vehicles", "campaigns", "media_types"
        ],
        organization_id: str | None = None,
        vehicle_ids: list[str] | None = None,
        media_type_ids: list[str] | None = None,
    ) -> list[dict[str, str]]:
        """Connects to the Mediatool API and retrieves data for the specified endpoint.

        Args:
            endpoint (Literal["organizations", "media_entries", "vehicles", "campaigns",
                "media_types"]): The API endpoint to fetch data from.
            organization_id (str, optional): Organization ID. Defaults to None.
            vehicle_ids (list[str]): List of organization IDs. Defaults to None.
            media_type_ids (list[str]): List of media type IDs. Defaults to None.

        Returns:
            list[dict[str, str]]: A list of records containing the retrieved data.
        """
        if endpoint == "organizations":
            return self._get_organizations(self.user_id)

        if endpoint == "media_entries":
            return self._get_media_entries(organization_id=organization_id)

        if endpoint == "vehicles":
            return self._get_vehicles(vehicle_ids=vehicle_ids)

        if endpoint == "campaigns":
            return self._get_campaigns(organization_id=organization_id)

        if endpoint == "media_types":
            return self._get_media_types(media_type_ids=media_type_ids)
        return None

    def fetch_and_transform(
        self,
        endpoint: Literal[
            "organizations", "media_entries", "vehicles", "campaigns", "media_types"
        ],
        organization_id: str | None = None,
        vehicle_ids: list[str] | None = None,
        media_type_ids: list[str] | None = None,
        columns: list[str] | None = None,
        if_empty: str = "warn",
        add_endpoint_suffix: bool = True,
    ) -> pd.DataFrame:
        """Pandas Data Frame with the data in the Response object and metadata.

        Args:
            endpoint (Literal["organizations", "media_entries", "vehicles", "campaigns",
                "media_types"]): The API endpoint to fetch data from.
            organization_id (str, optional): Organization ID. Defaults to None.
            vehicle_ids (list[str]): List of organization IDs. Defaults to None.
            media_type_ids (list[str]): List of media type IDs. Defaults to None.
            columns (list[str], optional): If provided, a list of column names to
                include in the DataFrame.By default, all columns will be included.
                Defaults to None.
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn
            add_endpoint_suffix (bool, optional): If True, appends the endpoint name
                to column names in the format {column_name}_{endpoint} to ensure
                uniqueness in combined DataFrame from multiple endpoints.
                Defaults to True.

        Returns:
            pd.Dataframe: The response data as a Pandas Data Frame plus viadot metadata.
        """
        records = self._to_records(
            endpoint, organization_id, vehicle_ids, media_type_ids
        )

        data_frame = pd.DataFrame.from_dict(records)  # type: ignore

        if endpoint == "campaigns":
            data_frame.replace(
                to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
                value=["", ""],
                regex=True,
                inplace=True,
            )

        if add_endpoint_suffix:
            data_frame = data_frame.rename(
                columns={
                    column_name: f"{column_name}_{endpoint}"
                    for column_name in data_frame.columns
                }
            )

        if columns:
            if set(columns).issubset(set(data_frame.columns)):
                data_frame = data_frame[columns]
            elif not set(columns).issubset(set(data_frame.columns)):
                self.logger.error(
                    f"Columns '{', '.join(columns)}' are incorrect. "
                    + "Whole dictionary for 'mediaEntries' will be returned."
                )

        if data_frame.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info(
                "Successfully downloaded data from "
                + f"the Mediatool API ({endpoint})."
            )

        return data_frame

    @add_viadot_metadata_columns
    def to_df(
        self,
        organization_ids: list[str],
        media_entries_columns: list[str] | None = None,
        if_empty: Literal["warn"] | Literal["skip"] | Literal["fail"] = "warn",
    ) -> pd.DataFrame:
        """Fetches, transforms, and combines data from Mediatool API endpoints.

        Args:
            organization_ids (list[str]): List of organization IDs.
            media_entries_columns (list[str], optional): Columns to get from media
                entries. Defaults to None.
            if_empty (Literal[warn, skip, fail], optional): What to do if there is no
                data. Defaults to "warn".

        Raises:
            ValueError: Raised when no organizations are defined or an organization ID
            is not found in the organizations list.

        Returns:
            pd.DataFrame: DataFrame containing the combined data from the specified
            endpoints.
        """
        if not organization_ids:
            message = "'organization_ids' must be a non-empty list."
            raise ValueError(message)

        # first method ORGANIZATIONS
        df_organizations = self.fetch_and_transform(endpoint="organizations")

        list_of_organizations_df = []
        for organization_id in organization_ids:
            if organization_id in df_organizations["_id_organizations"].unique():
                self.logger.info(f"Downloading data for: {organization_id} ...")

                # extract media entries per organization
                df_media_entries = self.fetch_and_transform(
                    endpoint="media_entries",
                    organization_id=organization_id,
                    columns=media_entries_columns,
                    add_endpoint_suffix=False,
                )

                unique_vehicle_ids = df_media_entries["vehicleId"].unique()
                unique_media_type_ids = df_media_entries["mediaTypeId"].unique()

                # extract vehicles
                df_vehicles = self.fetch_and_transform(
                    endpoint="vehicles", vehicle_ids=unique_vehicle_ids
                )

                # extract campaigns
                df_campaigns = self.fetch_and_transform(
                    endpoint="campaigns",
                    organization_id=organization_id,
                )

                # extract media types
                df_media_types = self.fetch_and_transform(
                    endpoint="media_types",
                    media_type_ids=unique_media_type_ids,
                )

                # join media entries & organizations
                df_merged_entries_orgs = join_dfs(
                    df_left=df_media_entries,
                    df_right=df_organizations,
                    left_on="organizationId",
                    right_on="_id_organizations",
                    columns_from_right_df=[
                        "_id_organizations",
                        "name_organizations",
                        "abbreviation_organizations",
                    ],
                    how="left",
                )

                # join the previous merge & campaigns
                df_merged_campaigns = join_dfs(
                    df_left=df_merged_entries_orgs,
                    df_right=df_campaigns,
                    left_on="campaignId",
                    right_on="_id_campaigns",
                    columns_from_right_df=[
                        "_id_campaigns",
                        "name_campaigns",
                        "conventionalName_campaigns",
                    ],
                    how="left",
                )

                # join the previous merge & vehicles
                df_merged_vehicles = join_dfs(
                    df_left=df_merged_campaigns,
                    df_right=df_vehicles,
                    left_on="vehicleId",
                    right_on="_id_vehicles",
                    columns_from_right_df=["_id_vehicles", "name_vehicles"],
                    how="left",
                )

                # join the previous merge & media types
                df_merged_media_types = join_dfs(
                    df_left=df_merged_vehicles,
                    df_right=df_media_types,
                    left_on="mediaTypeId",
                    right_on="_id_media_types",
                    columns_from_right_df=["_id_media_types", "name_media_types"],
                    how="left",
                )

                list_of_organizations_df.append(df_merged_media_types)

            else:
                message = (
                    f"Organization - {organization_id} not found in organizations list."
                )
                raise ValueError(message)

        df_final = pd.concat(list_of_organizations_df)

        if df_final.empty:
            self._handle_if_empty(if_empty)

        return df_final



"""A module for working with Amazon S3 as a data source."""

from collections.abc import Iterable, Iterator
from datetime import datetime
import io
import os
from pathlib import Path
from typing import Any, Literal

from boto3.s3.transfer import TransferConfig


try:
    import awswrangler as wr
    import boto3
    import s3fs
except ModuleNotFoundError:
    msg = "Missing required modules to use RedshiftSpectrum source."
    raise ImportError(msg) from None

import pandas as pd
from pydantic import BaseModel, root_validator

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source


class S3Credentials(BaseModel):
    region_name: str  # The name of the AWS region.
    aws_access_key_id: str  # The AWS access key ID.
    aws_secret_access_key: str  # The AWS secret access key.
    profile_name: str | None = None  # The name of the IAM profile to use.

    @root_validator(pre=True)
    def is_configured(cls, credentials: dict) -> dict:  # noqa: N805
        """Validate credentials.

        Ensure that at least one of the
        following is provided:
        - profile_name and region_name
        - aws_access_key_id, aws_secret_access_key, and region_name
        """
        profile_name = credentials.get("profile_name")
        region_name = credentials.get("region_name")
        aws_access_key_id = credentials.get("aws_access_key_id")
        aws_secret_access_key = credentials.get("aws_secret_access_key")

        profile_credential = profile_name and region_name
        direct_credential = aws_access_key_id and aws_secret_access_key and region_name

        if not (profile_credential or direct_credential):
            msg = "Either `profile_name` and `region_name`, or `aws_access_key_id`, "
            msg += "`aws_secret_access_key`, and `region_name` must be specified."
            raise CredentialError(msg)
        return credentials


class S3(Source):
    def __init__(
        self,
        credentials: S3Credentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """A class for pulling data from and uploading to the Amazon S3.

        Args:
        credentials (S3Credentials, optional): Amazon S3 credentials.
            Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        """
        raw_creds = (
            credentials
            or get_source_credentials(config_key)
            or self._get_env_credentials()
        )
        validated_creds = dict(S3Credentials(**raw_creds))  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        if not self.credentials:
            self.logger.debug(
                "Credentials not specified. Falling back to `boto3` default credentials."
            )

        self.fs = s3fs.S3FileSystem(
            profile=self.credentials.get("profile_name"),
            region_name=self.credentials.get("region_name"),
            key=self.credentials.get("aws_access_key_id"),
            secret=self.credentials.get("aws_secret_access_key"),
        )

        self._session = None

    @property
    def session(self) -> boto3.session.Session:
        """A singleton-like property for initiating a session to the AWS."""
        if not self._session:
            self._session = boto3.session.Session(
                region_name=self.credentials.get("region_name"),
                profile_name=self.credentials.get("profile_name"),
                aws_access_key_id=self.credentials.get("aws_access_key_id"),
                aws_secret_access_key=self.credentials.get("aws_secret_access_key"),
            )
        return self._session

    def _get_env_credentials(self):
        return {
            "region_name": os.environ.get("AWS_DEFAULT_REGION"),
            "aws_access_key_id": os.environ.get("AWS_ACCESS_KEY_ID"),
            "aws_secret_access_key": os.environ.get("AWS_SECRET_ACCESS_KEY"),
        }

    def ls(self, path: str, suffix: str | None = None) -> list[str]:
        """Returns a list of objects in a provided path.

        Args:
            path (str): Path to a folder.
            suffix (Union[str, List[str], None]): Suffix or list of suffixes for
                filtering Amazon S3 keys. Defaults to None.
        """
        return wr.s3.list_objects(boto3_session=self.session, path=path, suffix=suffix)

    def exists(self, path: str) -> bool:
        """Check if an object exists in the Amazon S3.

        Args:
            path (str): The path to an object to check.

        Returns:
            bool: Whether the object exists.
        """
        if not path.startswith("s3://"):
            msg = "Path must be an AWS S3 URL ('s3://my/path')."
            raise ValueError(msg)

        # Note this only checks for files.
        file_exists = wr.s3.does_object_exist(boto3_session=self.session, path=path)

        if file_exists:
            return True

        # Use another method in case the path is a folder.
        client = self.session.client("s3")
        bucket = path.split("/")[2]
        path = str(Path(*path.rstrip("/").split("/")[3:]))

        response = client.list_objects_v2(Bucket=bucket, Prefix=path, Delimiter="/")

        folders_with_prefix: list[dict] = response.get("CommonPrefixes")
        if folders_with_prefix is None:
            folder_exists = False
        else:
            # This is because list_objects takes in `Prefix`, so eg. if there exists
            #  a path `a/b/abc` and we run `list_objects_v2(path=`a/b/a`)`,
            #  it would enlist `a/b/abc` as well.
            paths = [path["Prefix"].rstrip("/") for path in folders_with_prefix]
            folder_exists = path in paths
        return folder_exists

    def cp(self, from_path: str, to_path: str, recursive: bool = False) -> None:
        """Copies the contents of `from_path` to `to_path`.

        Args:
            from_path (str): The path (S3 URL) of the source directory.
            to_path (str): The path (S3 URL) of the target directory.
            recursive (bool, optional): Set this to true if working with directories.
                Defaults to False.

        Example:
            Copy files within two S3 locations:

            ```python
            from viadot.sources.s3 import S3

            s3 = S3()
            s3.cp(
                from_path="s3://bucket-name/folder_a/",
                to_path="s3://bucket-name/folder_b/",
                recursive=True
            )
        """
        self.fs.copy(path1=from_path, path2=to_path, recursive=recursive)

    def rm(self, path: str | list[str]) -> None:
        """Delete files under `path`.

        Args:
            path (list[str]): Path to a list of files or a directory
                to be removed. If the path is a directory, it will
                be removed recursively.

        Example:
        ```python
        from viadot.sources import S3

        s3 = S3()
        s3.rm(path=["file1.parquet"])
        ```
        """
        wr.s3.delete_objects(boto3_session=self.session, path=path)

    def from_df(
        self,
        df: pd.DataFrame,
        path: str,
        extension: Literal[".csv", ".parquet"] = ".parquet",
        **kwargs,
    ) -> None:
        """Upload a pandas `DataFrame` into Amazon S3 as a CSV or Parquet file.

        For a full list of available parameters, please refer to the official
        documentation:
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.to_parquet.html
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.to_csv.html

        Args:
            df (pd.DataFrame): The pandas DataFrame to upload.
            path (str): The destination path.
            extension (Literal[".csv", ".parquet"], optional): The file extension.
                Defaults to ".parquet".
        """
        if extension == ".parquet":
            wr.s3.to_parquet(
                boto3_session=self.session,
                df=df,
                path=path,
                **kwargs,
            )
        elif extension == ".csv":
            wr.s3.to_csv(
                boto3_session=self.session,
                df=df,
                path=path,
                **kwargs,
            )
        else:
            msg = "Only CSV and Parquet formats are supported."
            raise ValueError(msg)

    def to_df(
        self,
        paths: list[str],
        chunk_size: int | None = None,
        **kwargs,
    ) -> pd.DataFrame | Iterable[pd.DataFrame]:
        """Read a CSV or Parquet file into a pandas `DataFrame`.

        Args:
            paths (list[str]): A list of paths to Amazon S3 files. All files under the
                path must be of the same type.
            chunk_size (int, optional): Number of rows to include in each chunk.
                Defaults to None, ie. return all data as a single `DataFrame`.

        Returns:
            Union[pd.DataFrame, Iterable[pd.DataFrame]]: A pandas DataFrame
                or an iterable of pandas DataFrames.

        Example 1:
        ```python
        from viadot.sources import S3

        s3 = S3()
        paths = ["s3://{bucket}/file1.parquet", "s3://{bucket}/file2.parquet"]
        s3.to_df(paths=paths)
        ```

        Example 2:
        ```python
        from viadot.sources import S3

        s3 = S3()
        paths = ["s3://{bucket}/file1.parquet", "s3://{bucket}/file2.parquet"]
        dfs = s3.to_df(paths=paths, chunk_size=2)

        for df in dfs:
            print(df)
        ```
        """
        if chunk_size is None:
            # `chunked` expects either an integer or a boolean.
            chunk_size = False

        if paths[0].endswith(".csv"):
            df = wr.s3.read_csv(
                boto3_session=self.session, path=paths, chunksize=chunk_size, **kwargs
            )
        elif paths[0].endswith(".parquet"):
            df = wr.s3.read_parquet(
                boto3_session=self.session, path=paths, chunked=chunk_size, **kwargs
            )
        else:
            msg = "Only CSV and Parquet formats are supported."
            raise ValueError(msg)
        return df

    def upload(self, from_path: str, to_path: str) -> None:
        """Upload file(s) to S3.

        Args:
            from_path (str): Path to local file(s) to be uploaded.
            to_path (str): Path to the destination file/folder.
        """
        wr.s3.upload(boto3_session=self.session, local_file=from_path, path=to_path)

    def download(self, from_path: str, to_path: str) -> None:
        """Download file(s) from Amazon S3.

        Args:
            from_path (str): Path to file in Amazon S3.
            to_path (str): Path to local file(s) to be stored.
        """
        wr.s3.download(boto3_session=self.session, path=from_path, local_file=to_path)

    def get_page_iterator(
        self,
        bucket_name: str,
        directory_path: str,
        operation_name: str = "list_objects_v2",
        **kwargs,
    ) -> Iterator[dict[str, Any]]:
        """Returns an iterator to paginate through the objects in S3 bucket directory.

        This method uses the S3 paginator to list objects under a specified directory
        path in a given S3 bucket. It can accept additional optional parameters
        through **kwargs, which will be passed to the paginator.

        Args:
            bucket_name (str): The name of the S3 bucket.
            directory_path (str): The directory path (prefix) in the bucket to list
                objects from.
            operation_name (str): The operation name. This is the same name as
                the method name on the client. Defaults as "list_objects_v2".
            **kwargs: Additional arguments to pass to the paginator (optional).

        Returns:
            Iterator: An iterator to paginate through the S3 objects.
        """
        client = self.session.client("s3")
        paginator = client.get_paginator(operation_name=operation_name)

        return paginator.paginate(Bucket=bucket_name, Prefix=directory_path, **kwargs)

    def get_object_sizes(self, file_paths: str | list[str]) -> dict[str, int | None]:
        """Retrieve the sizes of specified S3 objects.

        Args:
            file_paths (str | list[str]): A single file path or a list of file paths
                in S3 bucket.

        Returns:
            dict[str, int]: A dictionary where the keys are file paths and the values
                are their corresponding sizes in bytes.
        """
        return wr.s3.size_objects(boto3_session=self.session, path=file_paths)

    def describe_objects(
        self,
        file_paths: str | list[str],
        last_modified_begin: datetime | None = None,
        last_modified_end: datetime | None = None,
    ) -> dict[str, dict[str, Any]]:
        """Describe S3 objects from a received S3 prefix or list of S3 objects paths.

        Args:
            file_paths (str | list[str]): S3 prefix (accepts Unix shell-style wildcards)
                (e.g. s3://bucket/prefix) or list of S3 objects paths
                (e.g. [s3://bucket/key0, s3://bucket/key1]).
            last_modified_begin (datetime | None, optional): Filter the S3 files by
                the Last modified date of the object. The filter is applied only after
                list all s3 files. Defaults to None.
            last_modified_end (datetime | None, optional): Filter the s3 files by
                the Last modified date of the object. The filter is applied only after
                list all s3 files. Defaults to None.

        Returns:
            dict[str, dict[str, Any]]: Return a dictionary of objects returned from
            head_objects where the key is the object path.

        Examples:
            >>> s3 = S3(credentials=aws_credentials)
                # Describe both objects
            >>> descs0 = s3.describe_objects(['s3://bucket/key0', 's3://bucket/key1'])
                # Describe all objects under the prefix
            >>> descs1 = s3.describe_objects('s3://bucket/prefix')
        """
        return wr.s3.describe_objects(
            boto3_session=self.session,
            path=file_paths,
            last_modified_begin=last_modified_begin,
            last_modified_end=last_modified_end,
        )

    def upload_file_object(
        self,
        file_object: io.BytesIO,
        bucket_name: str,
        path: str,
        config: dict | None = None,
    ) -> None:
        """Upload a file-like object to S3.

        The file-like object must be in binary mode.
        This is a managed transfer which will perform a multipart upload in multiple
        threads if necessary.

        Args:
            file_object (io.BytesIO): A file-like object to upload.
            bucket_name (str): The name of the bucket to upload to.
            path (str): The name of the key to upload to.
            config (dict, optional): Configuration for multipart upload.
                If None, a default config with a 25MB multipart threshold and 10 threads
                is used.
        """
        if config is None:
            config = TransferConfig(
                multipart_threshold=25 * 1024 * 1024,  # 25 MB
                multipart_chunksize=25 * 1024 * 1024,  # 25 MB
                max_concurrency=10,
                use_threads=True,
            )
        else:
            config = TransferConfig(**config)

        client = self.session.client("s3")

        if not hasattr(file_object, "read"):
            msg = "file_object must be a file-like object with a read() method."
            raise ValueError(msg)

        client.upload_fileobj(
            Fileobj=file_object, Bucket=bucket_name, Key=path, Config=config
        )



from unittest import mock

from pydantic import SecretStr
import pytest

from viadot.exceptions import CredentialError
from viadot.sources.cloud_for_customers import (
    CloudForCustomers,
    CloudForCustomersCredentials,
)


def test_is_configured_valid():
    credentials = {
        "username": "user@tenant.com",
        "password": SecretStr("password"),
    }
    validated_creds = CloudForCustomersCredentials.is_configured(credentials)
    assert validated_creds == credentials


def test_is_configured_invalid():
    credentials = {"username": "user@tenant.com"}
    with pytest.raises(CredentialError):
        CloudForCustomersCredentials.is_configured(credentials)


@pytest.fixture
def c4c():
    credentials = {
        "username": "user@tenant.com",
        "password": SecretStr("password"),
        "url": "https://example.com",
        "report_url": "https://example.com/report",
    }
    return CloudForCustomers(credentials=credentials)


def test_create_metadata_url(c4c):
    url = "https://example.com/service.svc/Entity"
    expected_metadata_url = "https://example.com/service.svc/$metadata?entityset=Entity"
    assert c4c.create_metadata_url(url) == expected_metadata_url


def test_get_entities(c4c):
    dirty_json = {"d": {"results": [{"key": "value"}]}}
    url = "https://example.com/service.svc/Entity"

    c4c.create_metadata_url = mock.Mock(
        return_value="https://example.com/service.svc/$metadata?entityset=Entity"
    )
    expected_entities = {"key": "new_key"}
    c4c.get_property_to_sap_label_dict = mock.Mock(return_value=expected_entities)

    assert c4c.get_entities(dirty_json, url) == [{"new_key": "value"}]


@mock.patch("viadot.sources.cloud_for_customers.requests.get")
def test_get_property_to_sap_label_dict(mocked_requests_get, c4c):
    mocked_requests_get.return_value.text = (
        """<Property Name="key" sap:label="Label"/>"""
    )
    column_mapping = c4c.get_property_to_sap_label_dict(
        url="https://example.com/metadata"
    )
    assert column_mapping == {"key": "Label"}


@mock.patch("viadot.sources.cloud_for_customers.requests.get")
def test_get_response(mocked_requests_get, c4c):
    mock_response = mock.Mock()
    mock_response.status_code = 200
    mock_response.json.return_value = {"d": {"results": [{"key": "value"}]}}
    mocked_requests_get.return_value = mock_response
    c4c.get_response = mocked_requests_get

    response = c4c.get_response(url="https://example.com/service.svc/Entity")
    assert response.ok
    assert response.json() == {"d": {"results": [{"key": "value"}]}}



"""Download data from Exchange Rates API and upload it to AWS Redshift Spectrum."""

from datetime import datetime
from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import (
    df_to_redshift_spectrum,
    exchange_rates_to_df,
)
from viadot.orchestration.prefect.tasks.exchange_rates import Currency


@flow(
    name="extract--exchange-rates-api--redshift_spectrum",
    description="Extract data from Exchange Rates API and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
def exchange_rates_api_to_redshift_spectrum(  # noqa: PLR0913
    to_path: str,
    schema_name: str,
    table: str,
    currency: Currency = "USD",
    start_date: str = datetime.today().strftime("%Y-%m-%d"),
    end_date: str = datetime.today().strftime("%Y-%m-%d"),
    symbols: list[str] | None = None,
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    compression: str | None = None,
    aws_config_key: str | None = None,
    aws_credentials_secret: str | None = None,
    exchange_rates_api_credentials_secret: str | None = None,
    exchange_rates_api_config_key: str | None = None,
) -> None:
    """Extract data from Exchange Rates API and load it into AWS Redshift Spectrum.

    Args:
        currency (Currency, optional): Base currency to which prices of searched
            currencies are related. Defaults to "USD".
        start_date (str, optional): Initial date for data search.
            Data range is start_date -> end_date,
            supported format 'yyyy-mm-dd'.
            Defaults to datetime.today().strftime("%Y-%m-%d").
        end_date (str, optional): See above.
            Defaults to datetime.today().strftime("%Y-%m-%d").
        symbols (list[str], optional): List of ISO codes of currencies for which
            exchange rates from base currency will be fetched. Defaults to
            ["USD","EUR","GBP","CHF","PLN","DKK","COP","CZK","SEK","NOK","ISK"].
        to_path (str): Path to a S3 folder where the table will be located. Defaults to
            None.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        if_exists (str, optional): 'overwrite' to recreate any possible existing table
            or 'append' to keep any possible existing table. Defaults to overwrite.
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Only takes effect if dataset=True. Defaults to None.
        compression (str, optional): Compression style (None, snappy, gzip, zstd).
        sep (str, optional): Field delimiter for the output file. Defaults to ','.
        aws_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        aws_credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        exchange_rates_api_credentials_secret (str, optional): The name of the secret
            storing Exchange Rates API API key. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        exchange_rates_api_config_key (str, optional): The key in the viadot config
            holding relevant credentials. Defaults to None.
    """
    df = exchange_rates_to_df(
        currency=currency,
        start_date=start_date,
        end_date=end_date,
        symbols=symbols,
        credentials_secret=exchange_rates_api_credentials_secret,
        config_key=exchange_rates_api_config_key,
    )
    df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        if_exists=if_exists,
        partition_cols=partition_cols,
        compression=compression,
        config_key=aws_config_key,
        credentials_secret=aws_credentials_secret,
    )



"""Task for interacting with SAP."""

import contextlib
from typing import Any

import pandas as pd
from prefect import task
from prefect.logging import get_run_logger


with contextlib.suppress(ImportError):
    from viadot.sources import SAPRFC
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60 * 3)
def sap_rfc_to_df(  # noqa: PLR0913
    rfc_unique_id: list[str] | None = None,
    query: str | None = None,
    sep: str | None = "♔",
    func: str | None = None,
    replacement: str = "-",
    rfc_total_col_width_character_limit: int = 400,
    tests: dict[str, Any] | None = None,
    credentials_secret: str | None = None,
    credentials: dict[str, Any] | None = None,
    config_key: str | None = None,
    dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
    dynamic_date_format: str = "%Y%m%d",
    dynamic_date_timezone: str = "UTC",
) -> pd.DataFrame:
    """A task for querying SAP with SQL using the RFC protocol.

    Note that only a very limited subset of SQL is supported:
    - aliases
    - where clauses combined using the AND operator
    - limit & offset

    Unsupported:
    - aggregations
    - joins
    - subqueries
    - etc.

    Args:
        query (str): The query to be executed with pyRFC.
        sap_sep (str, optional): The separator to use when reading query results.
            If set to None, multiple options are automatically tried.
            Defaults to ♔.
        func (str, optional): SAP RFC function to use. Defaults to None.
        replacement (str, optional): In case of sep is on a columns, set up a new
            character to replace inside the string to avoid flow breakdowns. Defaults to
            "-".
        rfc_total_col_width_character_limit (int, optional): Number of characters by
            which query will be split in chunks in case of too many columns for RFC
            function. According to SAP documentation, the limit is 512 characters.
            However, we observed SAP raising an exception even on a slightly lower
            number of characters, so we add a safety margin. Defaults to 400.
        rfc_unique_id (list[str]):
            Reference columns to merge chunks DataFrames. These columns must to be
            unique.
        tests (dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from viadot.utils. Defaults to None.
        credentials_secret (str, optional): The name of the secret that stores SAP
            credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        credentials (dict[str, Any], optional): Credentials to SAP.
            Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        dynamic_date_symbols (list[str], optional): Symbols used for dynamic date
            handling. Defaults to ["<<", ">>"].
        dynamic_date_format (str, optional): Format used for dynamic date parsing.
            Defaults to "%Y%m%d".
        dynamic_date_timezone (str, optional): Timezone used for dynamic date
            processing. Defaults to "UTC".

    Examples:
        sap_rfc_to_df(
            ...
            rfc_unique_id=["VBELN", "LPRIO"],
            ...
        )
    """
    if not (credentials_secret or credentials or config_key):
        raise MissingSourceCredentialsError

    if query is None:
        msg = "Please provide the query."
        raise ValueError(msg)
    logger = get_run_logger()
    logger.warning("If the column/set are not unique the table will be malformed.")

    credentials = credentials or get_credentials(credentials_secret)

    sap = SAPRFC(
        sep=sep,
        replacement=replacement,
        credentials=credentials,
        func=func,
        rfc_total_col_width_character_limit=rfc_total_col_width_character_limit,
        rfc_unique_id=rfc_unique_id,
        config_key=config_key,
    )

    query = sap._parse_dates(
        query=query,
        dynamic_date_symbols=dynamic_date_symbols,
        dynamic_date_format=dynamic_date_format,
        dynamic_date_timezone=dynamic_date_timezone,
    )

    sap.query(query)
    logger.info("Downloading data from SAP to a DataFrame...")
    logger.debug(f"Running query: \n{sap.sql}.")

    df = sap.to_df(tests=tests)

    if not df.empty:
        logger.info("Data has been downloaded successfully.")
    elif df.empty:
        logger.warn("Task finished but NO data was downloaded.")
    return df



"""Flow for downloading data from Business Core API to a Parquet file."""

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks.business_core import business_core_to_df
from viadot.orchestration.prefect.tasks.task_utils import df_to_parquet


@flow(
    name="extract--businesscore--parquet",
    description="Extract data from Business Core API and load it into Parquet file",
    retries=1,
    retry_delay_seconds=60,
)
def business_core_to_parquet(
    path: str | None = None,
    url: str | None = None,
    filters: dict[str, Any] | None = None,
    credentials_secret: str | None = None,
    config_key: str | None = None,
    if_empty: str = "skip",
    if_exists: Literal["append", "replace", "skip"] = "replace",
    verify: bool = True,
) -> None:
    """Download data from Business Core API to a Parquet file.

    Args:
        path (str, required): Path where to save the Parquet file. Defaults to None.
        url (str, required): Base url to the view in Business Core API.
            Defaults to None.
        filters (dict[str, Any], optional): Filters in form of dictionary.
            Available filters: 'BucketCount', 'BucketNo', 'FromDate', 'ToDate'.
            Defaults to None.
        credentials_secret (str, optional): The name of the secret that stores Business
            Core credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        if_empty (str, optional): What to do if output DataFrame is empty.
            Defaults to "skip".
        if_exists (Literal["append", "replace", "skip"], optional):
            What to do if the table exists. Defaults to "replace".
        verify (bool, optional): Whether or not verify certificates while
            connecting to an API. Defaults to True.
    """
    df = business_core_to_df(
        url=url,
        credentials_secret=credentials_secret,
        config_key=config_key,
        filters=filters,
        if_empty=if_empty,
        verify=verify,
    )
    return df_to_parquet(
        df=df,
        path=path,
        if_exists=if_exists,
    )



"""Test ECB source."""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import requests

from viadot.sources.ecb import ECBExchangeRates


@pytest.fixture
def sample_ecb_response():
    """Load sample ECB API response for testing."""
    response_path = Path(__file__).parent / "test_ecb_response.xml"
    with response_path.open(encoding="utf-8") as f:
        return f.read()


@pytest.fixture
def ecb_instance():
    """Create ECBExchangeRates instance."""
    return ECBExchangeRates()


def test_init(ecb_instance):
    """Test ECBExchangeRates initialization."""
    assert ecb_instance.credentials is None
    assert (
        ecb_instance.URL
        == "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-daily.xml"
    )


@patch("viadot.sources.ecb.handle_api_response")
def test_fetch_data_success(
    mock_handle_api_response, ecb_instance, sample_ecb_response
):
    """Test successful data fetching."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    # Call fetch_data (no url parameter needed)
    data = ecb_instance.fetch_data()

    # Verify the data was returned
    assert data == sample_ecb_response

    # Verify handle_api_response was called correctly with hardcoded URL
    mock_handle_api_response.assert_called_once()
    call_args = mock_handle_api_response.call_args
    assert call_args[1]["url"] == ecb_instance.URL
    assert call_args[1]["params"] is None
    assert call_args[1]["method"] == "GET"


@patch("viadot.sources.ecb.handle_api_response")
def test_fetch_data_request_exception(mock_handle_api_response, ecb_instance):
    """Test fetch_data with request exception."""
    from viadot.exceptions import APIError

    mock_handle_api_response.side_effect = requests.exceptions.RequestException(
        "Connection failed"
    )

    with pytest.raises(APIError):
        ecb_instance.fetch_data()


def test_parse_xml(ecb_instance, sample_ecb_response):
    """Test XML parsing."""
    df = ecb_instance._parse_xml(sample_ecb_response)

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "time" in df.columns
    assert "currency" in df.columns
    assert "rate" in df.columns

    # Check that we have data
    assert len(df) > 0

    # Check that time is consistent
    assert df["time"].nunique() == 1
    assert df["time"].iloc[0] == "2025-12-30"

    # Check that currency values are strings
    assert df["currency"].dtype == "object"

    # Check that rate values are numeric
    assert pd.api.types.is_numeric_dtype(df["rate"])


def test_parse_xml_invalid_xml(ecb_instance):
    """Test parse_xml with invalid XML."""
    # Use XML that doesn't match ECB structure
    # This will parse successfully but fail structure validation
    invalid_xml = "<invalid>xml</invalid>"
    with pytest.raises(ValueError, match="Could not find time attribute"):
        ecb_instance._parse_xml(invalid_xml)


def test_parse_xml_missing_time(ecb_instance):
    """Test parse_xml with XML missing time attribute."""
    xml_without_time = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube currency='USD' rate='1.1757'/>
        </Cube>
    </gesmes:Envelope>"""
    with pytest.raises(ValueError, match="Could not find time attribute"):
        ecb_instance._parse_xml(xml_without_time)


def test_parse_xml_empty_data(ecb_instance):
    """Test parse_xml with XML containing no exchange rates."""
    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""
    df = ecb_instance._parse_xml(xml_empty)
    assert df.empty
    assert list(df.columns) == ["time", "currency", "rate"]


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_success(mock_handle_api_response, ecb_instance, sample_ecb_response):
    """Test successful to_df conversion."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    # Call to_df (no url parameter needed)
    df = ecb_instance.to_df()

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "time" in df.columns
    assert "currency" in df.columns
    assert "rate" in df.columns


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_empty_result_warn(mock_handle_api_response, ecb_instance):
    """Test to_df with empty result and warn if_empty."""
    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = xml_empty
    mock_handle_api_response.return_value = mock_response

    with patch("viadot.sources.base.logger.warning") as mock_warning:
        df = ecb_instance.to_df(if_empty="warn")

        assert df.empty

        assert mock_warning.call_count == 3
        mock_warning.assert_any_call("No exchange rate data found in XML response.")
        mock_warning.assert_any_call("No exchange rates found in the response.")
        mock_warning.assert_any_call("The query produced no data.")


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_empty_result_skip(mock_handle_api_response, ecb_instance):
    """Test to_df with empty result and skip if_empty."""
    from viadot.signals import SKIP

    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = xml_empty
    mock_handle_api_response.return_value = mock_response

    with pytest.raises(SKIP):
        ecb_instance.to_df(if_empty="skip")


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_empty_result_fail(mock_handle_api_response, ecb_instance):
    """Test to_df with empty result and fail if_empty."""
    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = xml_empty
    mock_handle_api_response.return_value = mock_response

    with pytest.raises(ValueError, match="The query produced no data"):
        ecb_instance.to_df(if_empty="fail")


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_with_tests_validation(
    mock_handle_api_response, ecb_instance, sample_ecb_response
):
    """Test to_df with tests validation."""
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    test_config = {"column_tests": {"currency": {"not_null": True}}}

    with patch("viadot.sources.ecb.validate") as mock_validate:
        df = ecb_instance.to_df(tests=test_config)

        mock_validate.assert_called_once()
        args, kwargs = mock_validate.call_args
        assert kwargs["df"] is df
        assert kwargs["tests"] == test_config


@patch("viadot.sources.ecb.handle_api_response")
def test_full_workflow(mock_handle_api_response, ecb_instance, sample_ecb_response):
    """Test complete workflow from fetch to DataFrame."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    # Fetch and convert to DataFrame (no url parameter needed)
    df = ecb_instance.to_df()

    # Verify results
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "time" in df.columns
    assert "currency" in df.columns
    assert "rate" in df.columns

    # Verify we have multiple currencies
    assert df["currency"].nunique() > 1



"""'test_sftp.py'."""

from io import BytesIO, StringIO
import json

import pandas as pd
import pytest

from viadot.exceptions import CredentialError
from viadot.sources import Sftp
from viadot.sources.sftp import SftpCredentials


variables = {
    "credentials": {
        "hostname": "",
        "username": "test_user",
        "password": "test_password",  # pragma: allowlist secret
        "port": 999,
        "rsa_key": "",
    },
}


class TestSftpCredentials:
    """Test SFTP Credentials Class."""

    @pytest.mark.basic
    def test_sftp_credentials(self):
        """Test SFTP credentials."""
        SftpCredentials(
            hostname=variables["credentials"]["hostname"],
            username=variables["credentials"]["username"],
            password=variables["credentials"]["password"],
            port=variables["credentials"]["port"],
            rsa_key=variables["credentials"]["rsa_key"],
        )


@pytest.mark.basic
def test_sftp_connector_initialization_without_credentials():
    """Test SFTP server without credentials."""
    with pytest.raises(CredentialError, match="Missing credentials."):
        Sftp(credentials=None)


@pytest.mark.connect
def test_get_connection_without_rsa_key(mocker):
    """Test `get_connection()` method without specifying the RSA key."""
    mock_transport = mocker.patch("viadot.sources.sftp.paramiko.Transport")
    mock_sftp_client = mocker.patch(
        "viadot.sources.sftp.paramiko.SFTPClient.from_transport"
    )

    connector = Sftp(credentials=variables["credentials"])
    connector.get_connection()

    mock_transport.assert_called_once_with((variables["credentials"]["hostname"], 999))
    mock_transport().connect.assert_called_once_with(
        None, variables["credentials"]["username"], variables["credentials"]["password"]
    )
    mock_sftp_client.assert_called_once()


@pytest.mark.connect
def test_get_connection_with_rsa_key(mocker):
    """Test SFTP `get_connection` method with ras_key."""
    mock_ssh_client = mocker.patch("viadot.sources.sftp.paramiko.SSHClient")
    mock_ssh_instance = mock_ssh_client.return_value
    mock_ssh_connect = mocker.patch.object(mock_ssh_instance, "connect")
    mock_rsa_key = mocker.patch(
        "viadot.sources.sftp.paramiko.RSAKey.from_private_key",
        return_value=mocker.Mock(),
    )
    mock_transport = mocker.Mock()

    # Ensure the SSHClient's transport attribute isn't None
    mock_ssh_instance._transport = mock_transport

    dummy_rsa_key = "test_rsa_key"
    keyfile = StringIO(dummy_rsa_key)
    credentials = variables["credentials"]
    credentials["rsa_key"] = keyfile.getvalue()

    connector = Sftp(credentials=variables["credentials"])
    connector.get_connection()

    mock_rsa_key.assert_called_once()
    mock_ssh_connect.assert_called_once_with(
        "", username="test_user", pkey=mock_rsa_key.return_value
    )
    assert connector.conn is not None


@pytest.mark.functions
def test_to_df_with_csv(mocker):
    """Test SFTP `to_df` method with csv."""
    mock_get_file_object = mocker.patch.object(Sftp, "_get_file_object", autospec=True)
    mock_get_file_object.return_value = BytesIO(b"col1,col2\n1,2\n3,4\n")

    connector = Sftp(credentials=variables["credentials"])
    df = connector.to_df(file_name="test.csv", sep=",")

    assert isinstance(df, pd.DataFrame)
    assert df.shape == (2, 4)
    assert list(df.columns) == [
        "col1",
        "col2",
        "_viadot_source",
        "_viadot_downloaded_at_utc",
    ]


@pytest.mark.functions
def test_to_df_with_json(mocker):
    """Test SFTP `to_df` method with json."""
    mock_get_file_object = mocker.patch.object(Sftp, "_get_file_object", autospec=True)
    json_data = json.dumps({"col1": [1, 3], "col2": [2, 4]})
    mock_get_file_object.return_value = BytesIO(json_data.encode("utf-8"))

    connector = Sftp(credentials=variables["credentials"])
    df = connector.to_df(file_name="test.json")

    expected_df = pd.DataFrame({"col1": [1, 3], "col2": [2, 4]})
    expected_df["_viadot_source"] = "Sftp"
    expected_df["_viadot_downloaded_at_utc"] = pd.Timestamp.now()

    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) == [
        "col1",
        "col2",
        "_viadot_source",
        "_viadot_downloaded_at_utc",
    ]


@pytest.mark.functions
def test_to_df_unsupported_file_type(mocker):
    """Test raising ValueError for unsupported file types."""
    mocker.patch.object(Sftp, "_get_file_object", return_value=BytesIO(b"dummy data"))
    sftp = Sftp(credentials=variables["credentials"])

    with pytest.raises(ValueError, match="Unable to read file"):
        sftp.to_df(file_name="test.txt")


@pytest.mark.functions
def test_to_df_empty_dataframe_warn(mocker, caplog):
    """Test handling of empty DataFrame with 'warn' option."""
    mocker.patch.object(Sftp, "_get_file_object", return_value=BytesIO(b"column"))
    sftp = Sftp(credentials=variables["credentials"])

    with caplog.at_level("INFO"):
        sftp.to_df(file_name="test.csv", if_empty="warn")
    assert "The response does not contain any" in caplog.text


@pytest.mark.functions
def test_ls(mocker):
    """Test SFTP `_ls` method."""
    mock_sftp = mocker.MagicMock()

    mock_sftp.listdir_attr.side_effect = [
        [
            mocker.MagicMock(st_mode=0o40755, filename="folder_a"),
            mocker.MagicMock(st_mode=0o100644, filename="file1.txt"),
            mocker.MagicMock(st_mode=0o100644, filename="file2.txt"),
        ],
        [
            mocker.MagicMock(st_mode=0o40755, filename="folder_b"),
            mocker.MagicMock(st_mode=0o100644, filename="file3.txt"),
        ],
    ]

    sftp = Sftp(credentials=variables["credentials"])
    sftp.conn = mock_sftp

    files_list = sftp._ls(path=".", recursive=False)

    assert len(files_list) == 2
    assert files_list == ["file1.txt", "file2.txt"]


@pytest.mark.functions
def test_recursive_ls(mocker):
    """Test SFTP recursive `_ls` method."""
    mock_sftp = mocker.MagicMock()
    mock_sftp.listdir_attr.side_effect = [
        [
            mocker.MagicMock(st_mode=0o40755, filename="folder_a"),
            mocker.MagicMock(st_mode=0o100644, filename="file1.txt"),
            mocker.MagicMock(st_mode=0o100644, filename="file2.txt"),
        ],
        [
            mocker.MagicMock(st_mode=0o40755, filename="folder_b"),
            mocker.MagicMock(st_mode=0o100644, filename="file3.txt"),
        ],
    ]

    sftp = Sftp(credentials=variables["credentials"])
    sftp.conn = mock_sftp

    files_list = sftp._ls(path=".", recursive=True)

    assert len(files_list) == 3
    assert files_list == ["folder_a/file3.txt", "file1.txt", "file2.txt"]


@pytest.mark.functions
def test_get_files_list(mocker):
    """Test SFTP `get_files_list` method."""
    mock_list_directory = mocker.patch.object(
        Sftp, "_ls", return_value=["file1.txt", "file2.txt"]
    )

    connector = Sftp(credentials=variables["credentials"])
    files = connector.get_files_list(path="test_path", recursive=False)

    assert files == ["file1.txt", "file2.txt"]
    mock_list_directory.assert_called_once_with(path="test_path", recursive=False)


@pytest.mark.functions
def test_close_conn(mocker):
    """Test SFTP `_close_conn` method."""
    mock_conn = mocker.MagicMock()
    connector = Sftp(credentials=variables["credentials"])
    connector.conn = mock_conn
    connector._close_conn()

    mock_conn.close.assert_called_once()
    assert connector.conn is None



"""'test_customer_gauge.py'."""

from viadot.orchestration.prefect.flows import customer_gauge_to_adls


def test_sftp_to_adls(azure_key_vault_secret, adls_azure_key_vault_secret):
    """Test Customer Gauge API prefect flow."""
    state = customer_gauge_to_adls(
        azure_key_vault_secret=azure_key_vault_secret,
        unpack_by_field_reference_cols=["properties"],
        unpack_by_nested_dict_transformer=[],
        adls_path="raw/dyvenia_sandbox/customer_gauge/customer_gauge.parquet",
        adls_azure_key_vault_secret=adls_azure_key_vault_secret,
        adls_path_overwrite=True,
    )
    all_successful = all(s.type == "COMPLETED" for s in state)
    assert all_successful, "Not all tasks in the flow completed successfully."



"""Hubspot API connector."""

from datetime import datetime
import json
import re
from typing import Any

import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    cast_df_cols,
    handle_api_response,
)


class HubspotCredentials(BaseModel):
    """HubSpot API credentials."""

    token: str


class Hubspot(Source):
    """A class that connects and extracts data from Hubspot API.

    Documentation is available here:
        https://developers.hubspot.com/docs/api/crm/understanding-the-crm.

    Connector allows to pull data in two ways:
        - using base API for crm schemas as an endpoint
            (eg. "contacts", ""line_items", "deals", ...),
        - using full url as endpoint.
    """

    API_URL = "https://api.hubapi.com"

    def __init__(
        self,
        *args,
        credentials: HubspotCredentials | None = None,
        config_key: str = "hubspot",
        **kwargs,
    ):
        """Create an instance of Hubspot.

        Args:
            credentials (Optional[HubspotCredentials], optional): Hubspot credentials.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "hubspot".

        Examples:
            hubspot = Hubspot(
                credentials=credentials,
                config_key=config_key,
            )
            hubspot.call_api(
                method=api_method,
                endpoint=endpoint,
                campaign_ids=campaign_ids,
                contact_type=contact_type,
                filters=filters,
                properties=properties,
                nrows=nrows,
            )
            data_frame = hubspot.to_df()

        Raises:
            CredentialError: If credentials are not provided in local_config or
                directly as a parameter.
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(HubspotCredentials(**raw_creds))
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.full_dataset = None

    def _date_to_unix_millis(self, date: str) -> int:
        """Convert date from "yyyy-mm-dd" to Unix timestamp in milliseconds.

        Milliseconds since 1970-01-01 (UTC). For example, "2023-04-06" -> 1680739200000.

        Args:
            date (str): Input date in format "yyyy-mm-dd".

        Returns:
            int: Milliseconds since 1970-01-01 00:00:00 UTC.
        """
        return int(datetime.timestamp(datetime.strptime(date, "%Y-%m-%d")) * 1000)

    def _get_api_url(
        self,
        endpoint: str | None = None,
        filters: dict[str, Any] | None = None,
        properties: list[str] | None = None,
    ) -> str:
        """Generates full url for Hubspot API given filters and parameters.

        Args:
            endpoint (Optional[str], optional): API endpoint or full URL.
                If a relative endpoint is provided, it will be resolved against API_URL.
                Defaults to None.
            filters (Optional[Dict[str, Any]], optional): Filters defined for the API
                body in specific order. Defaults to None.
            properties (Optional[List[str]], optional): List of user-defined columns to
                be pulled from the API. Defaults to None.

        Returns:
            str: The final API URL.
        """
        if not endpoint:
            msg = "Endpoint must be provided."
            raise ValueError(msg)

        # Full URL passthrough
        if endpoint.startswith("http"):
            return endpoint

        # Normalize leading slash
        endpoint = endpoint.lstrip("/")

        # Some product families (e.g., hubdb) expect direct prefixing
        if endpoint.startswith("hubdb"):
            url = f"{self.API_URL}/{endpoint}"
        elif filters:
            url = f"{self.API_URL}/crm/v3/objects/{endpoint}/search/?limit=100&"
        else:
            url = f"{self.API_URL}/crm/v3/objects/{endpoint}/?limit=100&"

        if properties:
            url += f'properties={",".join(properties)}&'

        return url

    def _format_filters(
        self,
        filters: list[dict[str, Any]] | None,
    ) -> list[dict[str, Any]]:
        """API body (filters) conversion from a user defined to API language.

        Note: Right now only converts date to Unix Timestamp.

        Args:
            filters (Optional[List[Dict[str, Any]]]): List of filters in JSON format.

        Returns:
            List[Dict[str, Any]]: List of cleaned filters in JSON format.
        """
        if not filters:
            return []

        for item in filters:
            for subitem in item["filters"]:
                for key in list(subitem.keys()):
                    lookup = subitem[key]
                    regex = re.findall(r"\d+-\d+-\d+", lookup)
                    if regex:
                        regex = self._date_to_unix_millis(lookup)
                        subitem[key] = f"{regex}"

        return filters

    def _get_request_body(self, filters: list[dict[str, Any]]) -> str:
        """Clean the filters body and convert to a JSON-formatted string.

        Args:
            filters (List[Dict[str, Any]]): Filters dictionary that will be passed to
                Hubspot API. Defaults to {}.

        Example:
                    filters = {
                                "filters": [
                                    {
                                    "propertyName": "createdate",
                                    "operator": "BETWEEN",
                                    "highValue": "2023-03-27",
                                    "value": "2023-03-26"
                                    }
                                ]
                            }
                Operators between the min and max value are listed below:
                [IN, NOT_HAS_PROPERTY, LT, EQ, GT, NOT_IN, GTE, CONTAINS_TOKEN,
                    HAS_PROPERTY, LTE, NOT_CONTAINS_TOKEN, BETWEEN, NEQ]
                LT - Less than
                LTE - Less than or equal to
                GT - Greater than
                GTE - Greater than or equal to
                EQ - Equal to
                NEQ - Not equal to
                BETWEEN - Within the specified range. In your request, use key-value
                    pairs to set highValue and value. Refer to the example above.
                IN - Included within the specified list. This operator is
                    case-sensitive, so inputted values must be in lowercase.
                NOT_IN - Not included within the specified list
                HAS_PROPERTY - Has a value for the specified property
                NOT_HAS_PROPERTY - Doesn't have a value for the specified property
                CONTAINS_TOKEN - Contains a token. In your request, you can use
                    wildcards (*) to complete a partial search. For example, use the
                    value *@hubspot.com to retrieve contacts with a HubSpot email
                    address.
                NOT_CONTAINS_TOKEN  -Doesn't contain a token

        Returns:
            str: JSON string with the request body.
        """
        return json.dumps({"filterGroups": filters, "limit": 100})

    def _build_headers(self) -> dict[str, str]:
        return {
            "Authorization": f'Bearer {self.credentials["token"]}',
            "Content-Type": "application/json",
        }

    def _api_call(
        self,
        url: str | None = None,
        body: str | None = None,
        method: str | None = None,
    ) -> dict:
        """General method to connect to Hubspot API and generate the response.

        Args:
            url (Optional[str], optional): Hubspot API url. Defaults to None.
            body (Optional[str], optional): Filters that will be pushed to the API body.
                Defaults to None.
            method (Optional[str], optional): Method of the API call. Defaults to None.

        Returns:
            dict: API response in JSON format.
        """
        headers = self._build_headers()

        response = handle_api_response(
            url=url, headers=headers, data=body, method=method
        )

        return response.json()

    def _get_offset_from_response(
        self, api_response: dict[str, Any]
    ) -> tuple[str | None, Any | None]:
        """Assign offset type/value depending on keys in API response.

        Args:
            api_response (Dict[str, Any]): API response in JSON format.

        Returns:
            tuple[str | None, Any | None]: (offset_type, offset_value)
        """
        if "paging" in api_response:
            offset_type = "after"
            offset_value = api_response["paging"]["next"][f"{offset_type}"]

        elif "offset" in api_response:
            offset_type = "offset"
            offset_value = api_response["offset"]

        else:
            offset_type = None
            offset_value = None

        return (offset_type, offset_value)

    def _extract_items(self, response: dict[str, Any]) -> list[Any]:
        """Extract the list of items from various HubSpot response shapes."""
        # Prefer v3 "results"
        if "results" in response and isinstance(response["results"], list):
            return response["results"]
        # Fallback: find the first list-valued key (e.g., v1 'contacts')
        for _, value in response.items():
            if isinstance(value, list):
                return value
        return []

    def _fetch(
        self,
        endpoint: str | None = None,
        filters: list[dict[str, Any]] | None = None,
        properties: list[str] | None = None,
        nrows: int = 1000,
    ) -> list[Any]:
        """General method to connect to Hubspot API and generate the response.

        Args:
            endpoint (Optional[str], optional): API endpoint for an individual request.
                Defaults to None.
            filters (Optional[List[Dict[str, Any]]], optional): Filters defined for the
                API body in specific order. Defaults to None.

        Example:
                    filters=[
                        {
                            "filters": [
                                {
                                    "propertyName": "createdate",
                                    "operator": "BETWEEN",
                                    "highValue": "1642636800000",
                                    "value": "1641995200000",
                                },
                                {
                                    "propertyName": "email",
                                    "operator": "CONTAINS_TOKEN",
                                    "value": "*@xxxx.xx",
                                },
                            ]
                        }
                    ],
            properties (Optional[List[Any]], optional): List of user-defined columns to
                be pulled from the API. Defaults to None.
            nrows (int, optional): Max number of rows to pull during execution.
                Defaults to 1000.
        """
        full_dataset: list[Any] = []
        url = self._get_api_url(
            endpoint=endpoint,
            filters=filters,
            properties=properties,
        )
        if filters:
            filters_formatted = self._format_filters(filters)
            body = self._get_request_body(filters=filters_formatted)
            method = "POST"
            partition = self._api_call(url=url, body=body, method=method)
            full_dataset = self._extract_items(partition)

            while "paging" in partition and len(full_dataset) < nrows:
                body = json.loads(self._get_request_body(filters=filters_formatted))
                body["after"] = partition["paging"]["next"]["after"]
                partition = self._api_call(
                    url=url, body=json.dumps(body), method=method
                )
                full_dataset.extend(self._extract_items(partition))

        else:
            method = "GET"
            partition = self._api_call(url=url, method=method)
            full_dataset = self._extract_items(partition)

            offset_type, offset_value = self._get_offset_from_response(partition)

            while offset_value and len(full_dataset) < nrows:
                url = self._get_api_url(
                    endpoint=endpoint,
                    properties=properties,
                    filters=filters,
                )
                url += f"{offset_type}={offset_value}"

                partition = self._api_call(url=url, method=method)
                full_dataset.extend(self._extract_items(partition))

                offset_type, offset_value = self._get_offset_from_response(partition)

        return full_dataset

    def _fetch_contact_ids(
        self,
        campaign_ids: list[str],
        contact_type: str = "influencedContacts",
    ) -> list[dict[str, Any]]:
        """Fetch influenced contact IDs for multiple campaigns.

        Builds a DataFrame with: campaign_id, contact_id, contact_type.

        Args:
            campaign_ids (list[str]): Campaign IDs to query.
            contact_type (str): Label to attach to each contact (e.g., "influenced").
        """
        rows: list[dict[str, Any]] = []

        for campaign_id in campaign_ids:
            url = f"https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}/reports/contacts/{contact_type}"

            partition = self._api_call(url=url, method="GET")
            results = (partition or {}).get("results", [])
            for item in results:
                contact_id = item.get("id")
                if contact_id is not None:
                    rows.append(
                        {
                            "campaign_id": campaign_id,
                            "contact_id": contact_id,
                            "contact_type": contact_type,
                        }
                    )

        return rows

    def _get_campaign_metrics(
        self,
        campaign_ids: list[str],
    ) -> list[dict[str, Any]]:
        """Fetch metrics for multiple campaigns.

        For each campaign, calls:
        https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}/reports/metrics

        The response structure may vary by account. This method attaches the
        `campaign_id` and copies metric fields into the output rows.

        Args:
            campaign_ids (list[str]): Campaign IDs to query.
        """
        rows: list[dict[str, Any]] = []

        for campaign_id in campaign_ids:
            url = f"https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}/reports/metrics"
            partition = self._api_call(url=url, method="GET")
            metrics = partition or {}
            row: dict[str, Any] = {"campaign_id": campaign_id}
            for key in [
                "sessions",
                "newContactsFirstTouch",
                "newContactsLastTouch",
                "influencedContacts",
            ]:
                if key in metrics:
                    row[key] = metrics[key]
            rows.append(row)

        return rows

    def _get_campaign_budget_totals(
        self,
        campaign_ids: list[str],
    ) -> list[dict[str, Any]]:
        """Fetch budget totals for multiple campaigns.

        For each campaign, calls:
        https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}/budget/totals

        The response includes many fields (budgetItems/spendItems), but we only keep:
        - budgetTotal
        - remainingBudget
        - spendTotal
        along with campaign_id.

        Args:
            campaign_ids (list[str]): Campaign IDs to query.
        """
        rows: list[dict[str, Any]] = []

        for campaign_id in campaign_ids:
            url = f"https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}/budget/totals"
            data = self._api_call(url=url, method="GET") or {}
            row: dict[str, Any] = {
                "campaign_id": campaign_id,
                "budgetTotal": data.get("budgetTotal"),
                "remainingBudget": data.get("remainingBudget"),
                "spendTotal": data.get("spendTotal"),
                "currencyCode": data.get("currencyCode"),
            }
            rows.append(row)

        return rows

    def _get_campaign_details(self, campaign_ids: list[str]) -> list[dict[str, Any]]:
        """Fetch details for multiple campaigns.

        For each campaign, calls:
        https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}?properties=...
        The HubSpot response keeps requested fields inside the 'properties' object.

        Args:
            campaign_ids (list[str]): Campaign IDs to query.
        """
        props = [
            "hs_name",
            "hs_start_date",
            "hs_end_date",
            "hs_notes",
            "hs_owner",
        ]

        rows: list[dict[str, Any]] = []

        for campaign_id in campaign_ids:
            url = f"https://api.hubapi.com/marketing/v3/campaigns/{campaign_id}?properties={','.join(props)}"
            data = self._api_call(url=url, method="GET") or {}
            row: dict[str, Any] = {"campaign_id": campaign_id}
            properties_obj = data.get("properties") or {}
            for p in props:
                row[p] = properties_obj.get(p)
            rows.append(row)

        return rows

    def call_api(
        self,
        method: str | None = None,
        endpoint: str | None = None,
        campaign_ids: list[str] | None = None,
        contact_type: str | None = None,
        filters: list[dict[str, Any]] | None = None,
        properties: list[str] | None = None,
        nrows: int = 1000,
    ) -> list[dict[str, Any]]:
        """Dispatch a HubSpot API call and return results.

        Args:
            method (str | None): Logical method selector ("get_all_contacts",
                "get_campaign_metrics", "get_campaign_details",
                "get_campaign_budget_totals", "fetch_contact_ids",
                or None for generic fetch).
            endpoint (str | None): Endpoint or full URL for generic fetch.
            campaign_ids (list[str] | None): Campaign IDs for campaign-specific methods.
            contact_type (str | None): Contact type for "fetch_contact_ids".
            filters (list[dict[str, Any]] | None): The filters to apply to the request.
            properties (list[str] | None): The properties to include in the request.
            nrows (int): Maximum number of rows to fetch.
        """
        if method == "get_all_contacts":
            data = self._fetch(
                endpoint="https://api.hubapi.com/contacts/v1/lists/all/contacts/all"
            )
        elif method == "get_campaign_metrics":
            data = self._get_campaign_metrics(
                campaign_ids=campaign_ids,
            )
        elif method == "get_campaign_budget_totals":
            data = self._get_campaign_budget_totals(
                campaign_ids=campaign_ids,
            )
        elif method == "get_campaign_details":
            data = self._get_campaign_details(
                campaign_ids=campaign_ids,
            )
        elif method == "fetch_contact_ids":
            data = self._fetch_contact_ids(
                campaign_ids=campaign_ids,
                contact_type=contact_type,
            )
        else:
            data = self._fetch(
                endpoint=endpoint,
                filters=filters,
                properties=properties,
                nrows=nrows,
            )

        return data

    @add_viadot_metadata_columns
    def to_df(
        self,
        data: list[dict[str, Any]] | None = None,
        if_empty: str = "warn",
    ) -> pd.DataFrame:
        """Generate a pandas DataFrame with the data in the Response and metadata.

        Args:
            data (list[dict[str, Any]] | None): The data to convert to pandas DataFrame.
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn".

        Returns:
            pd.DataFrame: The response data as a pandas DataFrame plus viadot metadata.
        """
        super().to_df(if_empty=if_empty)

        data_frame = pd.json_normalize(data)

        # change all object columns to string
        data_frame = cast_df_cols(data_frame, types_to_convert=["object"])

        if data_frame.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info("Successfully downloaded data from the Hubspot API.")

        return data_frame



import contextlib

from prefect import flow
import pytest

from viadot.exceptions import TableDoesNotExistError
from viadot.utils import skip_test_on_missing_extra


try:
    from viadot.sources import Databricks
except ImportError:
    skip_test_on_missing_extra(source_name="Databricks", extra="databricks")

from viadot.orchestration.prefect.tasks import df_to_databricks


TEST_SCHEMA = "test_viadot_schema"
TEST_TABLE = "test"


@pytest.fixture(scope="session", autouse=True)
def databricks() -> Databricks:
    databricks = Databricks(config_key="databricks-qa-elt")
    databricks.create_schema(TEST_SCHEMA)

    yield databricks

    with contextlib.suppress(TableDoesNotExistError):
        databricks.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE)
    with contextlib.suppress(Exception):
        databricks.drop_schema(TEST_SCHEMA)

    databricks.session.stop()


def test_df_to_databricks(
    TEST_DF, databricks: Databricks, databricks_credentials_secret
):
    @flow
    def test_flow():
        return df_to_databricks(
            df=TEST_DF,
            schema=TEST_SCHEMA,
            table=TEST_TABLE,
            credentials_secret=databricks_credentials_secret,
        )

    assert not databricks._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)

    test_flow()

    assert databricks._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)

    databricks.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE)
